// This file is part of libigl, a simple c++ geometry processing library.
//
// Copyright (C) 2013, 2017 Alec Jacobson <alecjacobson@gmail.com>
//
// This Source Code Form is subject to the terms of the Mozilla Public License
// v. 2.0. If a copy of the MPL was not distributed with this file, You can
// obtain one at http://mozilla.org/MPL/2.0/.
#ifndef IGL_OPENGL_GL_H
#define IGL_OPENGL_GL_H

#ifdef IGL_OPENGL2_GL_H
#  error "igl/opengl2/gl.h already included"
#endif

// Always use this:
//     #include "gl.h"
// Instead of:
//     #include <OpenGL/gl3.h>
// or
//     #include <GL/gl.h>
//

#include <glad/gl.h>

#endif
